package ru.dbotthepony.kstarbound.world.entities

import ru.dbotthepony.kommons.util.IStruct2d
import ru.dbotthepony.kstarbound.defs.DamageNotification
import ru.dbotthepony.kstarbound.defs.Drawable
import ru.dbotthepony.kstarbound.json.builder.IStringSerializable
import ru.dbotthepony.kstarbound.math.vector.Vector2d
import ru.dbotthepony.kstarbound.network.packets.DamageNotificationPacket
import ru.dbotthepony.kstarbound.network.packets.DamageRequestPacket
import ru.dbotthepony.kstarbound.network.packets.HitRequestPacket
import ru.dbotthepony.kstarbound.world.Direction
import ru.dbotthepony.kstarbound.world.World

/**
 * Monsters, NPCs, Players
 */
abstract class ActorEntity : DynamicEntity() {
	final override val movement: ActorMovementController = ActorMovementController()
	abstract val statusController: StatusController

	override fun move(delta: Double) {
		statusController.applyMovementControls()
		super.move(delta)
	}

	// uint8_t
	enum class DamageBarType(override val jsonName: String) : IStringSerializable {
		DEFAULT("Default"), NONE("None"), SPECIAL("Special")
	}

	open val health: Double
		get() = statusController.resources["health"]!!.value
	open val maxHealth: Double
		get() = statusController.resources["health"]!!.maxValue!!

	abstract val damageBarType: DamageBarType

	override fun onJoinWorld(world: World<*, *>) {
		super.onJoinWorld(world)
		statusController.init(world)
	}

	override fun tick(delta: Double) {
		super.tick(delta)

		if (isLocal)
			statusController.tick(delta)
		else
			statusController.tickRemote(delta)
	}

	override fun hitOther(damage: HitRequestPacket) {
		super.hitOther(damage)
		statusController.notifyHitDealt(damage.target, damage.request)
	}

	override fun damagedOther(notification: DamageNotificationPacket) {
		super.damagedOther(notification)
		statusController.notifyDamageDealt(notification.notification)
	}

	enum class PortraitMode(override val jsonName: String) : IStringSerializable {
		HEAD("head"),
		BUST("bust"),
		FULL("full"),
		FULL_NEUTRAL("fullneutral"),
		FULL_NUDE("fullnude"),
		FULL_NEUTRAL_NUDE("fullneutralnude");
	}

	@Suppress("NAME_SHADOWING")
	fun toAbsolutePosition(pos: Vector2d): Vector2d {
		var pos = pos

		if (movement.facingDirection == Direction.LEFT) {
			pos = pos.copy(x = -pos.x)
		}

		if (movement.rotation != 0.0) {
			pos = pos.rotate(movement.rotation)
		}

		return movement.position + pos
	}

	open fun portrait(mode: PortraitMode): List<Drawable> {
		return emptyList()
	}
}
