package ru.dbotthepony.kstarbound.world.entities.behavior

import org.apache.logging.log4j.LogManager
import org.classdump.luna.Table
import org.classdump.luna.exec.CallPausedException
import org.classdump.luna.lib.CoroutineLib
import org.classdump.luna.runtime.Coroutine
import ru.dbotthepony.kstarbound.defs.actor.behavior.NodeOutput
import ru.dbotthepony.kstarbound.defs.actor.behavior.NodeParameter
import ru.dbotthepony.kstarbound.lua.userdata.BehaviorState

class ActionNode(val name: String, val parameters: Map<String, NodeParameter>, val outputs: Map<String, NodeOutput>) : AbstractBehaviorNode() {
	private var coroutine: Coroutine? = null

	override fun run(delta: Double, state: BehaviorState): Status {
		var coroutine = coroutine
		var firstTime = false

		if (coroutine == null) {
			firstTime = true

			val fn = state.functions[name] ?: throw RuntimeException("How? $name")

			coroutine = state.lua.call(CoroutineLib.create(), fn)[0] as Coroutine
		}

		try {
			val result = if (firstTime) {
				val parameters = state.blackboard.parameters(parameters, this)
				state.lua.call(CoroutineLib.resume(), coroutine, parameters, state.blackboard, this, delta)
			} else {
				state.lua.call(CoroutineLib.resume(), coroutine, delta)
			}

			val status = result[0] as Boolean

			if (result.size >= 2) {
				val second = result[1] as? Table

				if (second != null) {
					state.blackboard.setOutput(this, second)
				}
			}

			val isDead = state.lua.call(CoroutineLib.status(), coroutine)[0] == "dead"

			if (!status) {
				return Status.FAILURE
			} else if (isDead) {
				return Status.SUCCESS
			} else {
				return Status.RUNNING
			}
		} catch (err: CallPausedException) {
			LOGGER.error("Behavior ActionNode '$name' called blocking code, which initiated pause. This is not supported.")
			return Status.FAILURE
		}
	}

	override fun reset() {
		coroutine = null
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
